def tidy(input) :     
    '''tidy input, remove EOL, remove empty lines
       input : list of file lines 
    ''' 
    output = []

    for l in input : 
        l  = l.strip(' \n')   # tidy end of lines 
        if len(l) == 0 :      # strip empty lines 
            continue 
        if l.find("RETURN") != -1 : 
            continue
        output.append(l)
    return output

def removeContinuationSymbols(input) : 
    '''remove continuation symbols from input
       input : list of file lines'''

    output = []
    ml     = '' # merged line 
    # fine line continuations
    for l in input : 
        ai = l.find('&') 
        
        if ai != -1 : 
            l = l.replace('&','')
            ml = ml+l
        else : 
            if len(ml) == 0 :
                ml = l
            else :
                ml = ml+l
            output.append(ml)
            ml = ''

    return output


def removeComments(input) : 
    '''remove comment lines'''
    output = [] 
    for l in input : 
        if l[0] == '!' :
            continue 
        else : 
            output.append(l)

    return output

def decodeFileLine(input) : 
    '''decode line
       input is a string of a mad8 line'''


    splitInput = input.split();    
    
    for i in range(0,len(splitInput)) : 
        splitInput[i] = splitInput[i].strip(',')

    d = dict() 

    if len(splitInput) == 1 : 
        pass
    elif len(splitInput) > 1 : 
        type = splitInput[1].strip(',');
        if type == 'LINE' :
            if len(splitInput) == 4 :
                input = input.replace(',',' ')
                splitInput = input.split()

            
            d = decodeLine(splitInput)
        elif type == 'INSTRUMENT' : 
            d = decodeNamed(splitInput)
        elif type == 'MONITOR' :
            d = decodeNamed(splitInput) 
        elif type == 'WIRE' : 
            d = decodeNamed(splitInput) 
        elif type == 'MARKER' : 
            d = decodeNamed(splitInput) 
        elif type == 'PROFILE' : 
            d = decodeNamed(splitInput) 
        elif type == 'LCAVITY' :
            d = decodeLcavity(splitInput)
        elif type == 'DRIFT' : 
            d = decodeDrift(splitInput) 
        elif type == 'SBEND' : 
            d = decodeSbend(splitInput)             
        elif type == 'QUADRUPOLE' : 
            d = decodeQuadrupole(splitInput)             
        elif type == 'SEXTUPOLE' : 
            d = decodeSextupole(splitInput)
        elif type == 'OCTUPOLE' : 
            d = decodeOctupole(splitInput)
        elif type == 'DECAPOLE' : 
            d = decodeDecapole(splitInput)
        elif type == 'MULTIPOLE' : 
            d = decodeMultipole(splitInput)
        elif type == 'VKICKER' : 
            d = decodeKicker(splitInput)
        elif type == 'HKICKER' : 
            d = decodeKicker(splitInput)
        elif type == 'RCOLLIMATOR' : 
            d = decodeCollimator(splitInput)
        elif type == 'ECOLLIMATOR' : 
            d = decodeCollimator(splitInput)
        else :
            if len(splitInput) == 2 : 
                d = decodeNamed(splitInput)
            else :
                d = decodeLcavity(splitInput)

    return d

def decodeLine(input) :
    d = decodeNameAndType(input) 

    input[3] = input[3].replace('(','')
    input[-1] = input[-1].replace(')','')    
    d['LINE'] = input[3:]    

        
    return d

def decodeNamed(input) :
    d = decodeNameAndType(input) 
    return d

def decodeLcavity(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeDrift(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeSbend(input) : 
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d    

def decodeQuadrupole(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeSextupole(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeOctupole(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeDecapole(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeMultipole(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeKicker(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeCollimator(input) :
    d = decodeNameAndType(input) 
    for t in input[2:] : 
        [key,val] = splitKeyValue(t)
        d[key] = val
    return d

def decodeNameAndType(input) : 
    name = input[0].strip(':')
    type = input[1]
    d = dict()
    d['name'] = name
    d['type'] = type
    return d 

def splitKeyValue(t) : 
    st = t.split('=')
    key   = st[0]
    value = float(st[1])

    return [key,value]



import pymad8.Saveline as _Saveline

class Mad8Converter :

    def __init__(self, fileName, ngenerate = 1000) :

        #Import .saveline
        self.saveline = _Saveline.Loader(fileName)
        self.fileName = fileName
        self.ngenerate = ngenerate

        #Create a new .gmad that includes the three converted files, how to handle beam options?

    def convert (self) :
        #Split the saveline into three files
        self._saveFile()

        #Convert the files into a gmad format
        components = 'components_' + self.fileName.replace('saveline', 'gmad')
        sequences = 'sequences_' + self.fileName.replace('saveline', 'gmad')
        samplers = 'samplers_' + self.fileName.replace('saveline', 'gmad')

        self._formatMad82Gmad('components_' + self.fileName, components)
        self._formatMad82Gmad('sequences_' + self.fileName, sequences)
        self._formatMad82Gmad('samplers_' + self.fileName, samplers)

        container = file(self.fileName.replace('saveline', 'gmad'), 'w')

        #Check how options and beam should be handled?
        container.write('include ' + components + ';\n'
                        'include ' + sequences + ';\n'
                        'include options.gmad;\n'
                        'include beam.gmad;\n'
                        'include ' + samplers + ';\n'
                        'option ngenerate=' + str(self.ngenerate) + ';')

        container.close()

    def _saveFile(self) :

        sequences = file('sequences_' + self.fileName, 'w')
        components = file('components_' + self.fileName, 'w')
        samplers = file('samplers_' + self.fileName, 'w')

        for e in self.saveline.elementList :
            self._removeConflictingGMADKeywords(self.saveline.elementDict[e], e)

        # if file exists overwrite, how?
        for i in self.saveline.sequences :
            sequences.write(i + '\n')
        for c in self.saveline.components :
            components.write(c + '\n')
        for s in self.saveline.samplers :
            samplers.write(s + '\n')

        sequences.close()
        components.close()
        samplers.close()

    def _removeConflictingGMADKeywords(self, elements, e) :

        if e == 'DUMP' :
            e = 'PMUD'
        temp = ''
        wasLine = False

        for element in elements :
            if element == 'LINE' :
                wasLine = True
                if type(element) == str :
                    temp += element + '=('
            elif element == 'MULTIPOLE' :
                temp = 'MARKER; #'

            else :
                if type(element) == list :
                    for prop in element :
                        if 'APERTURE=' in prop : #Need to remove APERTURE
                            continue
                        if prop == 'DUMP' : #Need to rename DUMP as conflicting keyword.
                            prop = 'PMUD'
                        self.saveline.samplers.append('sample, range = ' + prop)
                        temp += prop + ', '

                    temp = temp[:-2]

                    if wasLine :
                        temp += ')'
                        self.saveline.sequences.append(e + ': ' + temp)


                if type(element) == str :
                        temp += element + ', '

                elif not wasLine :
                    self.saveline.components.append(e + ': ' + temp)
                else :
                    wasLine = False

    def _formatMad82Gmad(self, mad8FileName, gmadFileName, removeAperture = True, removeMultipole = True) :
        i = open(mad8FileName,'r')
        o = open(gmadFileName,'w')

        # line for analysis
        lta = str()
        # boolian analysis or not
        ba  = False

        for l in i :
            # check if continued line
            l = l.strip('\n ')
            t = l.split()
            # if last token is an "&" then join the line
            if len(t) > 0 :
                if t[0] == '!' and t[-1] != '&' :
                    lta = l
                    ba  = False
                    o.write(lta+'\n')
                    continue
                elif l[-1] == '&' :
                    lta = lta+l[0:-1]
                    ba  = False
                elif len(lta) != 0 :
                    lta = lta+l[0:]
                    ba  = True
                else :
                    lta = l
                    ba  = True
            else :
                lta = l
                o.write(lta+'\n')
                continue

            if ba :
    #            print 'l  >',l
    #            print 'lta>',lta

                # lower case
                lta = lta.lower()
                # append semicolon
                lta = lta+';'

                # remove mad specific commands (should be saveline so no
                # problems but...)
                if lta.find('plot')     != -1 : lta = ''
                if lta.find('assign')   != -1 : lta = ''
                if lta.find('envelope') != -1 : lta = ''
                if lta.find('twiss')    != -1 : lta = ''
                if lta.find('survey')   != -1 : lta = ''
                if lta.find('title')    != -1 : lta = ''
                if lta.find('option')   != -1 : lta = ''
                if lta.find('optics')   != -1 : lta = ''
                if lta.find('btrns')    != -1 : lta = ''
                if lta.find('print')    != -1 : lta = ''
                if lta.find('return')   != -1 : lta = ''
                if lta.find('saveline') != -1 : lta = ''
                if lta.find('#') != -1 : lta = '!'

                # check for 4 letter commands
                if lta.find('sben') != -1 and lta.find('sbend') == -1 :
                    lta = lta.replace('sben','sbend')
                if lta.find('drif') != -1 and lta.find('drift') == -1 :
                    lta = lta.replace('drif','drift')
                if lta.find('sext') != -1 and lta.find('sextupole') == -1 :
                    lta = lta.replace('sext','sextupole')
                if lta.find('quad') != -1 and lta.find('quadrupole') == -1 :
                    lta = lta.replace('quad','quadrupole')
                if lta.find('octu') != -1 and lta.find('octupole') == -1 :
                    lta = lta.replace('octu','octupole')
                if lta.find('mark') != -1 and lta.find('marker') == -1 :
                    lta = lta.replace('mark','marker')
                if lta.find('moni') != -1 and lta.find('monitor') == -1 :
                    lta = lta.replace('moni','monitor')

                # change unsuported types
                lta = lta.replace('monitor','marker');
                lta = lta.replace('wire','marker');
                lta = lta.replace('prof','marker');

                # remove aperture parameters
                iap = lta.find('aperture')
                if removeAperture and (iap != -1) :
                    lta = lta[0:iap-2]+'; ! removed aperture'

                # remove multipoles
                if removeMultipole and (lta.find('multipole') != -1) :
                    lta = '! removed multipole'+lta
                o.write(lta+'\n')
                # clean up
                ba = False
                lta = str()

        i.close()
        o.close()
